@@ -1,5 +1,4 @@
 # -*- coding: utf-8 -*-
-import traceback
 import logging
 import os, gzip
 
@@ -11,8 +10,6 @@ from pypos3d.pftk.StructuredAttribut import GenericTransform, NodeInput, PoserMa
 from pypos3d.pftk.GeomCustom import GeomCustom
 from pypos3d.pftk.PoserMeshed import PoserProp, Camera, Light
 from pypos3d.pftk.Figure import Figure
-from pypos3d.wftk import WFBasic
-# package: pypos3d.pftk
 
 # 
 #  Path Match between a referenced path of Poser and a real path.
@@ -42,20 +39,59 @@ class PathMatch(object):
     def setExist(self, _exist):
         self._exist = _exist
 
+
+class GTLink():
+  ''' Thid class holds the links (parent, descendant) between channels. '''
+  def __init__(self, qualifiedName : str, gt=None):
+    self.qname = qualifiedName
+    self.gt = gt
+    self.parents = {}
+    self.desc = {}
+    #print('Adding GTLink({:s} as {:s})'.format(qualifiedName, 'direct' if gt else 'fwd'))
+    
+  def getParents(self):
+    locset = set(self.parents.values())
+    for p in self.parents.values():
+      subset = set(p.getParents())
+      locset |= subset
+      
+    return list(locset)
+      
+  def getDesc(self):
+    locset = set(self.desc.values())
+    for p in self.desc.values():
+      subset = set(p.getDesc())
+      locset |= subset
+      
+    return list(locset)
+
 # 
 class PoserFile(object):
   ''' Poser File internal representation.
   The aim of this class is to represent all kind of Poser files.
   But it works fine with : Characters (.cr2), Scenes (.pz3) and Props (.pp2)
   '''
-  # 
-  def __init__(self, fn):
+  
+  # Set of channel types that can be computed
+  __COMPUTABLETRANSF = { PoserToken.E_targetGeom,  PoserToken.E_valueParm, PoserToken.E_deformerPropChan, PoserToken.E_geomChan, \
+                         PoserToken.E_rotateX, PoserToken.E_rotateY, PoserToken.E_rotateZ,
+                         PoserToken.E_propagatingScale, PoserToken.E_propagatingScaleX, PoserToken.E_propagatingScaleY, PoserToken.E_propagatingScaleZ,
+                         PoserToken.E_translateX, PoserToken.E_translateY, PoserToken.E_translateZ,
+                         PoserToken.E_scale, PoserToken.E_scaleX, PoserToken.E_scaleY, PoserToken.E_scaleZ,
+                         PoserToken.E_visibility
+                         }
+  
+
+  def __init__(self, fn, createLinks=False):
     ''' Create a PoserFile from a file.
     
     Parameters
     ----------
     fn : str
       Filename to read
+    createLinks : Boolean, optional, default False
+      Create the internal mapping between channels and operators
+      Check referenced parts in each figure (parent, weld, figure Root)
       
     Raises
     ------
@@ -106,6 +142,9 @@ class PoserFile(object):
     self._illustrationParms = None
     self._renderDefaults = None
     self._faceRoom = None
+
+    # Channels link Dictionnary
+    self._gtlnk = {}
   
     #  Temporary Attributes 
     self._lastBodyIdx = 0
@@ -120,14 +159,23 @@ class PoserFile(object):
     isZipped = isCompressed(ft)
 
     logging.info("File[%s] : opening", fn)
-    rin =  gzip.open(fn, 'rb') if isZipped else open(fn, 'rb')
+    rin =  gzip.open(fn, 'rt', errors='replace') if isZipped else open(fn, 'rt', errors='replace')
     pfr = PoserFileParser(rin, PZ3_FNAT)
-    pfr.getToken()
-    if pfr.isLeftBracket():
+    code,*_ = pfr.getLine()
+    if code==PoserFileParser.TT_LEFTBACKET:
       self.read(pfr)
-
+#     except:
+#       exc_type, exc_value, exc_tb = sys.exc_info()
+#       traceback.print_exception(exc_type, exc_value, exc_tb)
+#       print('Exception in PoserFile.__init__:'+str(pfr.lineno()))
+           
     rin.close()
     logging.info("File[%s] : %d lines read", fn,  pfr.lineno())
+    
+    if createLinks:
+      self.__createLinks(True)
+      self.checkActorLinks()
+      
     #  if WFBasic.PYPOS3D_TRACE: print ('File({0:s} - Read Error {1:s}'.format(fn, str(e.args)))
       
 
@@ -161,58 +209,55 @@ class PoserFile(object):
   # Because it is the mean it uses to recognize a structured attribute.
   #    
   def read(self, st):
-    nextLine = True
-    st.getToken()
-    while st.ttype != PoserFileParser.TT_EOF:
-      if st.ttype == PoserFileParser.TT_EOL:
-        st.getToken()
-
-      if st.ttype == PoserFileParser.TT_WORD:
-        if st.isRightBracket():
-          #  End of current structured attribute
-          break
-
-        cn = st.sval
+    code,cn,rw = st.getLine()
+    
+    while (code!=PoserFileParser.TT_EOF) and (code!=PoserFileParser.TT_RIGHTBRACKET):
 
+      if code== PoserFileParser.TT_WORD:
         try:
-          vc = Lang[st.sval]
+          vc = Lang[cn]
           #  Known word
-          if vc.isStructured:
-            st.getToNextLine()
-            nom = "" if st.ttype == PoserFileParser.TT_EOL else st.sval
+          if vc==PoserToken.E_figure:
+            if self._lastFigure:
+              self._lastFigure.read(st)
+              # Should finish the current figure definition
+              logging.info("Line[%d]: figure [%s] defined", st.lineno(), self._lastFigure.getPrintName())
+            else:
+              # Figure part outside any delcared figure
+              #  We're not reading a CR2 or PP2 or PZ3 file, but
+              #  probably a Pose file ... nothing to do
+              logging.warning("Line[%d]: Figure block found without figure", st.lineno())
+              
+          elif vc.isStructured:
+            nom = rw # "" if st.ttype == PoserFileParser.TT_EOL else st.sval
             sta = create(vc, nom)
 
             #  Overload of object if definition part exists
             sta = self.addAttribut(sta)
 
             #  Read the opening bracket
-            st.getToken()
-            if st.isLeftBracket():
+            code,cn,rw = st.getLine()
+            if code==PoserFileParser.TT_LEFTBACKET:
               sta.read(st)
             else:
-              logging.error( "Line[%d]:  '{' is missing for %s", st.lineno(),  cn)
+              logging.error("Line[%d]:  '{' is missing for %s", st.lineno(), cn)
           else:
             if vc.isDirect:
-              st.getToNextLine()
+              #st.getToNextLine()
               logging.error("Line[%d]: setDirect() illegal at PoserFile level %s", st.lineno(), cn)
             else:
               # Read before add
-              sa = create(vc, st.sval)
-              sa.read(st)
+              sa = create(vc, cn)
+              sa.read(st, rw)
               self.addAttribut(sa)
 
-            nextLine = False
+            #nextLine = False
         except KeyError: # if vc == None: --> Unknow word
           logging.info("Line[%d] - Unknown word:%s", st.lineno() , cn)
 
-      if nextLine:
-        st.skipToNextLine()
-      else:
-        nextLine = True
-
       #  Get next token
-      st.getToken()
-
+      code,cn,rw = st.getLine()
+      # Wend
   # 
   # Add an object at highest level of the Poser File.
   #    
@@ -299,7 +344,7 @@ class PoserFile(object):
         #  We are in the declarative part of "free" props
         self._lstProp.append(pp)
       else:
-        pp = self._lastFigure.addReadProp(pp)
+        pp = self._lastFigure.addReadActor(pp)
       return pp
 
     elif pot==PoserToken.E_light:
@@ -320,23 +365,16 @@ class PoserFile(object):
       self._lstCamera.append(cam)
       return cam
 
-    elif pot==PoserToken.E_figure:
-      fig2 = self.getFigure(self._lastBodyIdx)
-      if fig2:
-        fig2.setFigureDescr(po)
-      else:
-        #  We're not reading a CR2 or PP2 or PZ3 file, but
-        #  probably a Pose file ... nothing to do
-        logging.warning("Figure chapter found without figure declaration")
-      return po
-
+    # elif pot==PoserToken.E_figure: Managed in read() as of V0.2
     # Manage other Props : hairProp, controlProp, Magnets
     elif pot==PoserToken.E_controlProp or \
          pot==PoserToken.E_baseProp or \
          pot==PoserToken.E_hairProp or \
          pot==PoserToken.E_magnetDeformerProp or \
          pot==PoserToken.E_sphereZoneProp or \
-         pot==PoserToken.E_curveProp:
+         pot==PoserToken.E_curveProp or \
+         pot==PoserToken.E_waveDeformerProp or\
+         pot==PoserToken.E_coneForceFieldProp:
       pp = po
       pp.setPoserFile(self)
       if pp.getName() == "FocusDistanceControl":
@@ -351,7 +389,7 @@ class PoserFile(object):
         #  We are in the declarative part of "free" props
         self._lstProp.append(pp)
       else:
-        pp = self._lastFigure.addReadProp(pp)
+        pp = self._lastFigure.addReadActor(pp)
 
       return pp
 
@@ -420,8 +458,34 @@ class PoserFile(object):
   # @return
   #    
   def getLstFigProp(self):
-    return [ n for f in self.getLstFigure() for n in f.getDescendant() ] + \
-           [ pp.getName() for pp in self.getLstProp() ]
+    return [ pp for pp in self.getLstProp() ] +\
+           [ n for f in self.getLstFigure() for n in f.getActors() ] 
+   
+   
+  def findFigure(self, frf):
+    for fig in self._lstFigure:
+      if frf.getValue() == fig.getFigResFile().getValue():
+        return fig
+    return None
+
+  def findActor(self, actorName):
+    for fig in self._lstFigure:
+      for act in fig.getActors():
+        if act.getName() == actorName:
+          return act
+    return None
+
+  def findMeshedObject(self, actorPropName):
+    for fig in self._lstFigure:
+      for a in fig.getActors():
+        if actorPropName == a.getName():
+          return a
+
+    for pp in self._lstProp + self._lstLight + self._lstCamera:
+      if actorPropName == pp.getName():
+        return pp
+
+    return None
 
   # 
   # Return the list of PoserMeshedObject (actor, prop, hairprop) 
@@ -432,11 +496,10 @@ class PoserFile(object):
   def findAllMeshedObject(self, actorPropName=None):
     if actorPropName:
       return [ pp for pp in self._lstProp + self._lstLight+self._lstCamera if actorPropName == pp.getName() ] + \
-            [ part for fig in self._lstFigure for part in fig.getActors() + fig.getProps() if actorPropName == part.getName() ]
+            [ part for fig in self._lstFigure for part in fig.getActors() if actorPropName == part.getName() ]
     else:
       return self._lstProp + self._lstLight + self._lstCamera + \
-        [ a for fig in self._lstFigure for a in fig.getActors() ] + \
-        [ p for fig in self._lstFigure for p in fig.getProps()  ] \
+        [ a for fig in self._lstFigure for a in fig.getActors() ] 
 
   # 
   def writeFile(self, fn):
@@ -538,6 +601,36 @@ class PoserFile(object):
   def getChannel(self, groupName, channelName):
     po = self.findMeshedObject(groupName)
     return po.getGenericTransform(channelName) if po else None
+  
+  # Return the ancestors channel list of the channel given by channelName in
+  # groupName actor.
+  # 
+  # @param figureName
+  #          Unused because of welknown CrossWalk problems
+  # @param groupName
+  # @param channelName
+  # @return The list of GenericTransform
+  def getChannelAncestor(self, figureName, groupName, channelName):
+    lstact = self.findAllMeshedObject(groupName)   
+    if lstact and (len(lstact)==1):
+      po = lstact[0]
+      
+      channel = po.getChannel(channelName) # next(attr for attr in po.getChannels() if attr.getName()==channelName)
+      
+      if not self._gtlnk:
+        self.__createLinks(True)
+
+      qn = channel.getPoserMeshedObject().getName() + '.' + channel.getName()
+      try:
+        gtl = self._gtlnk[qn]
+        return [ channel, ] + [ descgtl.gt for descgtl in gtl.getParents() ]
+      except KeyError:
+        logging.warning("Channel[%d] not found", qn)
+        return []
+
+
+    logging.warning(figureName + "." + groupName + " Not Found")
+    return []
 
   # 
   # Return the k factor of the first frame of the channel given by groupName of
@@ -552,8 +645,9 @@ class PoserFile(object):
   def getFactor(self, figureName, groupName, channelName):
     po = self.findActor(groupName)
     if po:
-      for attr in po.getChannels().getLstAttr():
-        if ((attr.getPoserType() == PoserToken.E_targetGeom) or (attr.getPoserType() == PoserToken.E_valueParm)) and (attr.getName() == channelName):
+      for attr in po.getChannels():
+        if (attr.getName()==channelName) and \
+           (attr.getPoserType() in PoserFile.__COMPUTABLETRANSF):
           finalFactor = attr.getKeysFactor(0)
 
           #  Get Recursively controling channels
@@ -562,36 +656,38 @@ class PoserFile(object):
             if (groupName != vod.getGroupName()) or (channelName != vod.getChannelName()):
               # Get the real factor of the master
               f = self.getFactor(vod.getFigureName(), vod.getGroupName(), vod.getChannelName())
-              cr = vod.getControlRatio()
-              finalFactor += f * cr
+              finalFactor = vod.calc(finalFactor, f)
                     
           return finalFactor
+        
+      logging.warning("No Channel [%s] for [%s.%s]", channelName, figureName, groupName)
 
     logging.warning(figureName + "." + groupName + " Not Found")
     return 0.0
 
-  def calcChannelAncestors(self, gt, lstgt, lstMissing):
-    vodTab = gt.getVOD()
-    for op in vodTab:
-      found = False
-      for gl in lstgt:
-        if gl.getPoserMeshedObject().getName()==op.getGroupName() and gl.getName()==op.getChannelName():
-          # Found : Already in the list
-          found = True
-          break
-
-      if not found:
-        newgt = self.getChannel(op.getGroupName(), op.getChannelName())
-        if newgt:
-          lstgt.append(newgt)
-          self.calcChannelAncestors(newgt, lstgt, lstMissing)
-        else:
-          lstMissing.append(op.getGroupName() + '.' + op.getChannelName())
-
-  def getChannelAncestors(self, gt, lstMissing):
-    lstgt = [ ]
-    self.calcChannelAncestors(gt, lstgt, lstMissing)
-    return lstgt
+# 2020-11-06 Unused?
+#   def calcChannelAncestors(self, gt, lstgt, lstMissing):
+#     vodTab = gt.getVOD()
+#     for op in vodTab:
+#       found = False
+#       for gl in lstgt:
+#         if gl.getPoserMeshedObject().getName()==op.getGroupName() and gl.getName()==op.getChannelName():
+#           # Found : Already in the list
+#           found = True
+#           break
+# 
+#       if not found:
+#         newgt = self.getChannel(op.getGroupName(), op.getChannelName())
+#         if newgt:
+#           lstgt.append(newgt)
+#           self.calcChannelAncestors(newgt, lstgt, lstMissing)
+#         else:
+#           lstMissing.append(op.getGroupName() + '.' + op.getChannelName())
+# 2020-11-06 Unused?
+#   def getChannelAncestors(self, gt, lstMissing):
+#     lstgt = [ ]
+#     self.calcChannelAncestors(gt, lstgt, lstMissing)
+#     return lstgt
 
   def getDescendant(self, bodyIdx, firstName=None):
     f = self.getFigure(bodyIdx)
@@ -666,7 +762,7 @@ class PoserFile(object):
       ficpath = frf.getPath(poserRootDir)
       self.addToSet(poserRootDir, lstres, ficpath)
 
-      for pa in fig.getActors()+fig.getProps():
+      for pa in fig.getActors():
         self.refPoserMesh(poserRootDir, lstres, pa)
 
       for mat in fig.getLstMaterial():
@@ -707,38 +803,33 @@ class PoserFile(object):
         #idxRelPos = 3 if (bodyIdx > 9) else 2
 
         for po in lstActor:
-          #searchName = groupName if(groupName.rfind(':')==len(groupName) - idxRelPos) else groupName + ":" + str(bodyIdx)
+          for attr in po.getChannels():
+            if attr.getPoserType()==PoserToken.E_targetGeom and GenericTransform.concerned(attr.getName(), setTargetMorph):
+              finalFactor = 0.0
+              for vod in attr.getVOD():
+                # Get the real factor of the master
+                f = self.getFactor(vod.getFigureName(), vod.getGroupName(), vod.getChannelName())
+                # Take into account the operator kind
+                finalFactor = vod.calc(finalFactor, f)
 
-          #po = self.findActor(searchName)
-          #if po:
-            for attr in po.getChannels().getLstAttr():
-              if attr.getPoserType()==PoserToken.E_targetGeom and GenericTransform.concerned(attr.getName(), setTargetMorph):
-                finalFactor = 0.0
-                for vod in attr.getVOD():
-                  # Get the real factor of the master
-                  f = self.getFactor(vod.getFigureName(), vod.getGroupName(), vod.getChannelName())
-                  # Take into account the Control Ratio (deltaAddDelta)
-                  finalFactor += f*vod.getControlRatio()
-
-                # Deltas dltAttr = attr.getDeltas();
-                finalFactor += attr.getKeysFactor(0)
-                logging.info("Cleaning[%s] in %s f=%g", attr.getName(), po.getName(), finalFactor)
-                attr.removeDeltas()
+              # Deltas dltAttr = attr.getDeltas();
+              finalFactor += attr.getKeysFactor(0)
+              logging.info("Cleaning[%s] in %s f=%g", attr.getName(), po.getName(), finalFactor)
+              attr.removeDeltas()
       else:
         ret = C_ERROR
 
     else: # pp should be a PoserProp
       if not pp: return C_ERROR
       
-      for attr in pp.getChannels()._lstAttr:
+      for attr in pp.getChannels():
         if attr.getPoserType()==PoserToken.E_targetGeom and GenericTransform.concerned(attr.getName(), setTargetMorph):
           finalFactor = 0.0
           for vod in attr.getVOD():
             # Get the real factor of the master
             f = self.getFactor(vod.getFigureName(), vod.getGroupName(), vod.getChannelName())
-            # Take into account the Control Ratio (deltaAddDelta)
-            cr = vod.getControlRatio()
-            finalFactor += f * cr
+            # Take into account the operator kind
+            finalFactor = vod.calc(finalFactor, f)
 
           finalFactor += attr.getKeysFactor(0)
 
@@ -773,8 +864,6 @@ class PoserFile(object):
       lstgeom = [ ]
       for pa in lstactors:
         # Check if the name of the actor vs the name of the meshed part
-        #pa = self.findActor(pn)
-        #if pa:
         if pa.getGeomType()==PoserConst.GT_GLOBAL_OBJFILE:
           # geomHandlerGeom 13 hip
           eg = body.extractSortGeom(pa.getGeomGroupName(), pa.getGeomType())
@@ -795,61 +884,188 @@ class PoserFile(object):
     return lstgeom
 
   # Calculate the impact of a channel deletion. (Not recursive search)
-  # The first item is the concerned channel.
+  # The "first" item is the concerned channel.
   # @param channel
   def deleteChannelImpact(self, channel):
-    implst = [ channel, ]
-    for f in self.getLstFigure():
-      for pa in f.getActors():
-        pa.getChannelDescendant(channel, implst)
-      for pp in f.getProps():
-        pp.getChannelDescendant(channel, implst)
-    for pp in self.getLstProp():
-      pp.getChannelDescendant(channel, implst)
-    return implst
+    impset = { channel, }
+    
+    if self._gtlnk:
+      qn = channel.getPoserMeshedObject().getName() + '.' + channel.getName()
+      gtl = self._gtlnk[qn]
+      impset.update( descgtl.gt for descgtl in gtl.desc.values() )
+    else:  
+      for f in self.getLstFigure():
+        for pa in f.getActors():
+          pa.getChannelDescendant(channel, impset)
+#         for pp in f.getProps():
+#           pp.getChannelDescendant(channel, impset)
+      for pp in self.getLstProp():
+        pp.getChannelDescendant(channel, impset)
+        
+    return impset
 
-  # Calculate the list of channels that reference at any level the given channel. (recursive search)
+  # Calculate the set of channels that are referenced at any level by the given channel. (recursive search)
   # @param channel
+  # Raise KeyError when the channel is not known
+  def calcChannelParent(self, channel):
+    if not self._gtlnk:
+      self.__createLinks(True)
+
+    qn = channel.getPoserMeshedObject().getName() + '.' + channel.getName()
+    gtl = self._gtlnk[qn]
+
+    lstgtl = gtl.getParents()
+
+    return [ descgtl.gt for descgtl in lstgtl ]
+
+  # Calculate the set of channels that reference at any level the given channel. (recursive search)
+  # @param channel
+  # Raise KeyError when the channel is not known
   def calcChannelDesc(self, channel):
-    implst = [ ]
-    l = self.deleteChannelImpact(channel)
-    del l[0]
-    for gt in l:
-      if not gt in implst:
-        implst.append(gt)
-        lstsub = self.calcChannelDesc(gt)
-        for gt2 in lstsub:
-          if not gt2 in implst:
-            implst.append(gt2)
+    if not self._gtlnk:
+      self.__createLinks(True)
+
+    qn = channel.getPoserMeshedObject().getName() + '.' + channel.getName()
+    gtl = self._gtlnk[qn]
+
+    lstgtl = gtl.getDesc()
+
+    return [ descgtl.gt for descgtl in lstgtl ]
+    
+
+  def printChannelLinks(self, channel):
+    ''' Return a tuple of printable data for the channel (Qualified Name, nbDescDelta, DescChannelsNames, ParentChannelsNames) '''
+    if not self._gtlnk:
+      self.__createLinks(True)
+
+    if isinstance(channel, str): # Input as string as a qualified string
+      try:
+        gt = self._gtlnk[channel].gt
+      except KeyError:
+        # Channel does not exist
+        return (channel, -1, '', '')
+    else:
+      gt = channel
+      
+    # Compute channel dependencies
+    nbdeltas, labparent, labdesc = 0, '',''
+    try:
+      lparent = self.calcChannelParent(gt)
+      for gtdesc in [ g for g in lparent if g ]:
+        if labparent:
+          labparent += ', '          
+        labparent += gtdesc.getQualifiedName()
+      
+      ldesc = self.calcChannelDesc(gt)
+      for gtdesc in [ g for g in ldesc if g ]:
+        nbdeltas += len(gtdesc._dlt.deltaSet) if gtdesc._dlt and gtdesc.getPoserType()==PoserToken.E_targetGeom else 0
+        if labdesc:
+          labdesc += ', '
+        labdesc += gtdesc.getQualifiedName()
+      
+    except KeyError:
+      # Channel does not exist
+      return (channel, -1, '', '')
+    
+    return (channel.getQualifiedName(), nbdeltas, labdesc, labparent)
+
+  def __createLinks(self, printErrors=False):
+    ''' Create the link dictionnary of all channels in the file. '''
+    logging.info("Linking[%s]:Channels", self.getName())
+    self._gtlnk = {}
+
+    nbdeclfwd,nblinks = 0,0
+
+    for pmo in self.getLstFigProp():
+      for gt in pmo.getChannels():
+        qn = pmo.getName() + '.' + gt.getName()
+        # Check if the channel has already been seen else create it
+        try:
+          gtl = self._gtlnk[qn]
+
+          if gtl.gt==None: # If it was discovered before declaration
+            nbdeclfwd-=1
+            gtl.gt = gt
+
+        except KeyError:
+          gtl = GTLink(qn, gt)
+          self._gtlnk[qn] = gtl
+
+        for vop in gt.getVOD():
+          # Check valueOp syntax, the figure index seems to be optional
+          gn = vop.getGroupName()
+          
+          # Search for the channel as given by the valueOp
+          refqn = gn + '.' +  vop.getChannelName()
+          try:
+            refgtl = self._gtlnk[refqn]
+            
+          except KeyError:
+            p = gn.rfind(':')
+            if p<0:
+              refqn = gn + ':' + str(pmo.getIndex()) + '.' +  vop.getChannelName()
+              # Make a second search with a qualified name
+              try:
+                refgtl = self._gtlnk[refqn]
+              except KeyError:
+                refgtl = GTLink(refqn)
+                self._gtlnk[refqn] = refgtl
+                nbdeclfwd+=1
+            else:
+              refgtl = GTLink(refqn)
+              self._gtlnk[refqn] = refgtl
+              nbdeclfwd+=1
+
+          # The referenced channel is a parent for the current one
+          gtl.parents[refqn] = refgtl
+
+          # The current channel is a descendant for the channel named in the valueOp
+          refgtl.desc[qn] = gtl
+          nblinks+=1
+
+    logging.info("Link.end[%s]: NbChannels=%d, Links=%d, Unlinked=%d", self.getName(), len(self._gtlnk), nblinks, nbdeclfwd)
+    if printErrors and nbdeclfwd>0:
+      for k,gtl in self._gtlnk.items():
+        if not gtl.gt:
+          print('Not Declared Channel:{:s}'.format(k))
+  
+  def getGTLinks(self):
+    if not self._gtlnk:
+      self.__createLinks(True)
+    return self._gtlnk
+
+  def delGTLink(self, gtOrQn):
+    if not self._gtlnk: return
+    
+    qn = gtOrQn.getQualifiedName() if isinstance(gtOrQn, GenericTransform) else gtOrQn
+      
+    del self._gtlnk[qn]
+    for gtl in self._gtlnk.values():
+      gtl.desc.pop(qn, None)
 
-    return implst
 
   # Delete a channel in a PoserMeshedObject
   # @param fig
   #          Figure where is the channel (if the meshed object is an Actor)
   # @param lstImp
   #          Prepared impact list. If null will be calculated again.
-  def deleteChannel(self, channel, lstImp):
+  def deleteChannel(self, channel, lstImp=None):
     if lstImp == None:
       lstImp = self.deleteChannelImpact(channel)
-    fci = lstImp.pop(0)
-    pmo = fci.getPoserMeshedObject()
-    pmo.deleteChannel(fci)
+    
+    lstImp.remove(channel)
+    
+    # Remove calculated links
+    self.delGTLink(channel)
+    
+    pmo = channel.getPoserMeshedObject()
+    pmo.deleteChannel(channel)
     for ci in lstImp:
       ci.deleteChannelRef(channel)
+    
 
 
-  def findMeshedObject(self, actorPropName):
-    for fig in self._lstFigure:
-      for a in fig.getActors() + fig.getProps():
-        if actorPropName == a.getName():
-          return a
 
-    for pp in self._lstProp + self._lstLight + self._lstCamera:
-      if actorPropName == pp.getName():
-        return pp
-
-    return None
 
   def deleteFigureImpact(self, fig):
     logging.error("Not implemented yet")
@@ -921,59 +1137,17 @@ class PoserFile(object):
             pmo.setConformingTarget(None)
 
           # Delete any references in channels of actors of other figures and Props
-          for pgt in pmo.getChannels().getLstAttr():
+          for pgt in pmo.getChannels():
             pgt.deleteFigureRef(obj)
 
       except:
         res = C_FAIL
+        
+    # Remove calculated links (anyway, to be consistent)
+    self._gtlnk.clear()
 
     return res
 
-
-  def findFigure(self, frf):
-    for fig in self._lstFigure:
-      if frf.getValue() == fig.getFigResFile().getValue():
-        return fig
-    return None
-
-  def findActor(self, actorName):
-    for fig in self._lstFigure:
-      for act in fig.getActors():
-        if act.getName() == actorName:
-          return act
-    return None
-
-  # Return the ancestors channel list of the channel given by channelName in
-  # groupName actor.
-  # 
-  # @param figureName
-  #          Unused because of welknown CrossWalk problems
-  # @param groupName
-  # @param channelName
-  # @return The list of GenericTransform
-  def getChannelAncestor(self, figureName, groupName, channelName):
-    lstact = self.findAllMeshedObject(groupName)
-    reslst = [ ]
-    if lstact and (len(lstact)==1):
-      po = lstact[0]
-      
-      for attr in po.getChannels().getLstAttr():
-        if attr.getName() == channelName:
-          reslst.append(attr)
-          # Get Recursively controlling channels
-          for vod in attr.getVOD():
-            # Get other ancestors of the master
-            sublst = self.getChannelAncestor(vod.getFigureName(), vod.getGroupName(), vod.getChannelName())
-            # Add new found ancestors channels
-            for agt in sublst:
-              if not agt in reslst:
-                reslst.append(agt)
-
-          return reslst
-
-    logging.warning(figureName + "." + groupName + " Not Found")
-    return reslst
-
   # Change the name of referenced part (actor, prop, hairProp, controlProp)  
   # @param oldPartName
   # @param newPartName
@@ -981,3 +1155,42 @@ class PoserFile(object):
     for fig in self._lstFigure + self._lstProp:
       fig.changeReference(oldPartName, newPartName)
 
+
+
+
+  def checkActorLinks(self):
+    ''' Check all actors links in all figure and props '''
+
+    logging.info("Linking[%s]:Actors", self.getName())
+    nberrors=0
+    
+    for fig in self._lstFigure:
+      nberrors += fig.checkActorLinks()
+
+    for act in self._lstProp:
+      parentName = act.getParent()
+      
+      if parentName==PoserConst.C_UNIVERSE:
+        if self._universe:
+          continue
+        else:
+          nberrors+=1
+          logging.warn('Missing Declared Parent Actor: %s for %s', parentName, act.getName())        
+      else:
+        p = self.findActor(parentName)
+        if not p:
+          nberrors+=1
+          logging.warn('Missing Declared Parent Actor: %s for %s', act.getParent(), act.getName())
+
+    if nberrors:
+      logging.info("Linking[%s]:Actors with %d errors", self.getName(), nberrors)
+      
+    return C_OK if nberrors==0 else C_FAIL
+      
+      
+      
+      
+      
+      
+
+
