@@ -5,9 +5,11 @@ import logging
 
 from langutil import RemoveExt, C_FAIL, C_OK, C_ERROR
 from pypos3d.wftk.WFBasic import C_BAD_DELTAINDEX, C_BAD_DELTACOUNT, C_NODELTA, Vector3d
-from pypos3d.pftk.PoserBasic import PoserConst, PoserToken, PToken, buildRelPath, getRealPath, index, TAS, WBoolLine, create
+from pypos3d.wftk.PoserFileParser import PoserFileParser, ParsingErrorException
+from pypos3d.pftk.PoserBasic import PoserConst, PoserToken, PToken, buildRelPath, getRealPath, index, TAS, WBoolLine, create, Lang
 from pypos3d.pftk.GeomCustom import GeomCustom
-from pypos3d.pftk.StructuredAttribut import StructuredAttribut, Channels, AlternateGeom, Deltas, ValueOpDelta, ChannelMorphStatus, ChannelMorphStatusList, findNewDelta, filterLength, calcMapping_KDTree
+from pypos3d.pftk.StructuredAttribut import StructuredAttribut, AlternateGeom, Deltas, ValueOpDelta, ChannelMorphStatus, \
+    GenericTransform, ChannelMorphStatusList, findNewDelta, filterLength, calcMapping_KDTree
 
 
 #
@@ -54,10 +56,10 @@ class PoserMeshedObject(StructuredAttribut):
     #  Storage class of the mesh 
     self._geomType = PoserConst.GT_NONE
   
-    #  Geom Groupname when geomType is GLOBALself._OBJFILE 
+    #  Geom Groupname when geomType is GLOBAL_OBJFILE 
     self._geomGroupName = None
   
-    #  Geom filename when geomType is LOCALself._OBJFILE, formated like Poser with ':' or ""  
+    #  Geom filename when geomType is LOCAL_OBJFILE, formated like Poser with ':' or ""  
     self._geomFileName = None
   
     #  Base Geometry
@@ -67,8 +69,11 @@ class PoserMeshedObject(StructuredAttribut):
     #  FigureResFile with a path (relative) 
     self._FRF = None
   
-    #  Channels = list of targetGeom and so on - Never null 
-    self._channels = Channels()
+    #  Channels = Ordered Dictionary of Channels (only GenericTransform) - Never null 
+    #  { ChannelName in lower char : GenericTransform }
+    self._channels = { }
+    #  Store the group presentation of channels 
+    self._groups = None
   
     #  endPoint position 
     self._endPoint = None
@@ -100,7 +105,9 @@ class PoserMeshedObject(StructuredAttribut):
 
   def getPoserFile(self): return self._pf
 
-  def getChannels(self): return self._channels
+  def getChannels(self): 
+    ''' Return a View on the channels '''
+    return self._channels.values()
 
   def getGeomType(self): return self._geomType
 
@@ -124,7 +131,7 @@ class PoserMeshedObject(StructuredAttribut):
 
   def getFigure(self): return self._fig
 
-  # Return the list of alternate geom. Null if none (never null in Python)
+  # Return the list of alternate geom (never null in Python)
   def getAltGeomList(self): return self._lstAltGeom
 
   #    
@@ -151,7 +158,7 @@ class PoserMeshedObject(StructuredAttribut):
         ag.setPrintName(n)
         ag.setGeomFileName(relPath)
 
-      except IOError as ioex:
+      except IOError:
         ret = C_ERROR
 
    
@@ -198,31 +205,99 @@ class PoserMeshedObject(StructuredAttribut):
   # Return the alternate geom channel. Null if none
   #    
   def getAltGeomChannel(self):
-    for gt in self.getChannels().getLstAttr():
-      if (gt.getPoserType() == PoserToken.E_geomChan):
-        return gt
+    return next( (gt for gt in self.getChannels() if gt.getPoserType()==PoserToken.E_geomChan), None)
 
-    return None
 
   def getChannel(self, channelName):
-    channelName = channelName.lower()
-    for gt in self.getChannels().getLstAttr():
-      if gt.getName().lower() == channelName:
-        return gt
+    return self._channels.get(channelName.lower())
+  
+  # Find the GenericTransform that represents a TargetGeom and that has the
+  # given "channelName"
+  # @param channelName
+  #          Channel Name to look for.
+  # @return Null if not found.
+  def getTargetGeom(self, channelName):
+    gt = self._channels.get(channelName.lower())
+    return gt if gt and (gt.getPoserType()==PoserToken.E_targetGeom) else None
 
-    return None
+  # 
+  # Find the GenericTransform that represents the given "channelName" or PoserToken
+  # @param channelName
+  #          Channel Name to look for.
+  # @return Null if not found.
+  def getGenericTransform(self, channel:'Name Or PoserToken'):
+    return next( (gt for gt in self.getChannels() if gt.getPoserType()==channel), None) if isinstance(channel, PToken) \
+                  else self._channels.get(channel.lower())
+    
 
+  # 
+  # Update or create a targetGeom and create a dependency to the body homonyme valueParm (if bodyLinked=true)
+  #
+  # @param channelName
+  #    *          TargetGeom name
+  # @param targetFigureName
+  #    *          Referenced Figure Name
+  # @param srcWG
+  #    *          Concerned WaveGeom
+  # @param lstNewDeltas
+  #    *          List of deltas
+  # @param bodyIdx
+  #    *          Body index of the referenced figure (silly)
+  #    
+  #     updateOrCreate(String channelName, String targetFigureName, WaveGeom3d srcWG, lstNewDeltas, bodyIdx, bodyLinked)
+  # def updateOrCreate_0(self, channelName, targetFigureName, srcWG, lstNewDeltas, bodyIdx):
+  def updateOrCreate(self, channelName, targetFigureName, srcWG, setNewDeltas, bodyIdx, bodyLinked=True):
+    gt = self._channels.get(channelName.lower())
+    if not gt or (gt.getPoserType()!=PoserToken.E_targetGeom): #  Create the missing targetGeom
+      if bodyLinked:
+        gt = GenericTransform(PoserToken.E_targetGeom, channelName, targetFigureName, \
+                              self.getFigure().getRoot() if targetFigureName else None,\
+                              channelName if targetFigureName else None)
+                              #None if not targetFigureName else "BODY:" + str(bodyIdx),\
+                              #None if not targetFigureName else channelName)
+      else:
+        gt = GenericTransform(PoserToken.E_targetGeom, channelName)
+        #  Do not add a prefix to printable name (old P4 Stuff)
+        gt.setPrintName(channelName[3:] if (channelName.startswith("PBM") or channelName.startswith("FBM")) else channelName)
+        gt.addKeyFrame(0, 0.0)
+        #  Default keys are static=0
+        gt.setInterpStyleLocked(0)
 
+      self.addAttribut(gt)
+
+    gt.getDeltas().deltaSet = setNewDeltas # [ delta for delta in lstNewDeltas ] )
+    gt.setNumbDeltas(len(srcWG.coordList))
+    return gt
+
+  # 
+  # Update or create a valueParm and initializes it
+  # @param channelName ValueParm name
   def updateOrCreateVP(self, channelName, minVal=0.0, maxVal=1.0, applyLimits=True, isHidden=False):
-    return self.getChannels().updateOrCreateVP(channelName, minVal, maxVal, applyLimits, isHidden)
+    vp = self._channels.get(channelName.lower())
+    if not vp or (vp.getPoserType()!=PoserToken.E_valueParm):
+      #  Create the missing valueParm
+      vp = GenericTransform(PoserToken.E_valueParm, channelName)
+      #  Do not add a prefix to printable name (old P4 Stuff)
+      vp.setPrintName(channelName[3:] if (channelName.startswith("PBM") or channelName.startswith("FBM")) else channelName)
+      vp.addKeyFrame(0, 0.0)
+      #  Default keys are static=0
+      vp.setInterpStyleLocked(0)
+      self.addAttribut(vp)
+
+    vp.setMin(minVal)
+    vp.setMax(maxVal)
+    vp.setForceLimits(1 if applyLimits else 0)
+    vp.setHidden(isHidden)
+    return vp
+
 
   def CreateShaderNodeP(self, channelName, isHidden=False):
-    snp = self.getChannels().updateOrCreateVP(channelName, minVal=0.0, maxVal=1.0, applyLimits=True, isHidden=isHidden)
+    snp = self.updateOrCreateVP(channelName, minVal=0.0, maxVal=1.0, applyLimits=True, isHidden=isHidden)
     snp.setPoserType(PoserToken.E_shaderNodeParm)
     return snp
   
   def CreateDeformerP(self, channelName, zoneName, deformerName, isHidden=False):
-    snp = self.getChannels().updateOrCreateVP(channelName, minVal=-100.0, maxVal=100.0, applyLimits=True, isHidden=isHidden)
+    snp = self.updateOrCreateVP(channelName, minVal=-100.0, maxVal=100.0, applyLimits=True, isHidden=isHidden)
     snp.setPoserType(PoserToken.E_deformerPropChan)
 
     # Create Deformer Channel special attributs
@@ -236,24 +311,6 @@ class PoserMeshedObject(StructuredAttribut):
     return snp
   
 
-  def addAttribut(self, po):
-    if po.getPoserType()==PoserToken.E_geomCustom:
-      self._geomCustom = po
-      self._geomType = PoserConst.GT_INTERNAL
-    elif po.getPoserType()==PoserToken.E_channels:
-      self._channels = po
-      self._channels.setPoserMeshedObject(self)
-    elif po.getPoserType()==PoserToken.E_storageOffset:
-      # FIX20090607 : Storage Offset seen : shall not be kept in attribute list.
-      return
-    elif po.getPoserType()==PoserToken.E_alternateGeom:
-      self._lstAltGeom.append(po)
-      po.setNo(len(self._lstAltGeom))
-    elif po.getPoserType()==PoserToken.E_weightMap:
-      self._lstWMap.append(po)
-    else:
-      super(PoserMeshedObject, self).addAttribut(po)
-
   def getBaseMesh(self, poserRootDir):
     if self._geom == None:
       self._geom = self.getBaseGeomCustom(poserRootDir).getWaveGeom()
@@ -464,7 +521,7 @@ class PoserMeshedObject(StructuredAttribut):
 
       tabMapping = calcMapping_KDTree(srcWG, lstRefGeom, ropt.translation, ropt.maxDist)
       
-      channelsAttr = self.getChannels()
+      #channelsAttr = self.getChannels()
 
       # Foreach morph target ==> Exploit tabMapping for srcGC
       for channelName in setTargetMorph:
@@ -478,7 +535,7 @@ class PoserMeshedObject(StructuredAttribut):
           # Filter the result list to remove too short vectors
           lstFinalDeltas = filterLength(setNewDeltas, ropt.minVectLen)
 
-          channelsAttr.updateOrCreate(channelName, None, srcWG, lstFinalDeltas, 0)
+          self.updateOrCreate(channelName, None, srcWG, lstFinalDeltas, 0)
           logging.info("Deltas created for %s.%s : %d", srcGC.getName(), channelName, len(lstFinalDeltas))
         else:
           logging.warn("No Delta created for %s.%s", srcGC.getName(), channelName )
@@ -497,7 +554,7 @@ class PoserMeshedObject(StructuredAttribut):
 
     # Find the actor      
     srcActor = curPoserObject.findActor(self.getName())
-    channelsAttr = srcActor.getChannels()
+    #channelsAttr = srcActor.getChannels()
 
     # Extract and load alternate geoms
     lstAltG = self.getAltGeomList()
@@ -507,7 +564,7 @@ class PoserMeshedObject(StructuredAttribut):
       # GenericTransform altGeomUsageIndic[] = new GenericTransform[lstAltG.size() + 1];
 
       # Find the geomChan Channel (should always exists if the list of alt geom is not empty)
-      geomChan = channelsAttr.findAttribut(PoserToken.E_geomChan)
+      geomChan = self.getAltGeomChannel() # channelsAttr.findAttribut(PoserToken.E_geomChan)
       
       altno = 1
       for altg in lstAltG:
@@ -541,7 +598,7 @@ class PoserMeshedObject(StructuredAttribut):
             setFinalDeltas = filterLength(setNewDeltas, ropt.minVectLen)
 
             # Create or update the channel list
-            destGT = channelsAttr.updateOrCreate(altChannelName, self.getName(), srcWG, setFinalDeltas, srcActor.getFigure().getBodyIndex(), False)
+            destGT = self.updateOrCreate(altChannelName, self.getName(), srcWG, setFinalDeltas, srcActor.getFigure().getBodyIndex(), False)
             
             vod = ValueOpDelta(PoserToken.E_valueOpKey, srcActor.getFigure().getRefName(), srcActor.getName(), geomChan.getName(), 0.0)
 
@@ -551,7 +608,7 @@ class PoserMeshedObject(StructuredAttribut):
             destGT.addVOD(vod)
 
             # Multiply by the value of the copied channel
-            vodMul = ValueOpDelta(PoserToken.E_valueOpTimes, srcActor.getFigure().getName(), srcActor.getName(), channelName, 0.0)
+            vodMul = ValueOpDelta(PoserToken.E_valueOpTimes, srcActor.getFigure().getRefName(), srcActor.getName(), channelName, 0.0)
             destGT.addVOD(vodMul)
 
             destGT.setHidden(True)
@@ -560,30 +617,6 @@ class PoserMeshedObject(StructuredAttribut):
 
     return ret
 
-  # Find the GenericTransform that represents a TargetGeom and that has the
-  # given "channelName"
-  # @param channelName
-  #          Channel Name to look for.
-  # @return Null if not found.
-  def getTargetGeom(self, channelName):
-    for gt in self.getChannels().getLstAttr():
-      if gt.getPoserType() == PoserToken.E_targetGeom and gt.getName() == channelName:
-        return gt
-    return None
-
-  # 
-  # Find the GenericTransform that represents the given "channelName"
-  # @param channelName
-  #          Channel Name to look for.
-  # @return Null if not found.
-  def getGenericTransform(self, channel:'Name Or PoserToken'):
-    try:
-      return next(gt for gt in self.getChannels().getLstAttr() if gt.getPoserType()==channel) if isinstance(channel, PToken) \
-        else next(gt for gt in self.getChannels().getLstAttr() if gt.getName()==channel)
-      
-    except StopIteration:
-      return None
-    
     
   # 
   # Return the delta for a given channel and a given vertex no.
@@ -595,35 +628,51 @@ class PoserMeshedObject(StructuredAttribut):
   # @return Delta Object (Aka 3D vector)
   def findDelta(self, channelName, refNo):
     gt = self.getTargetGeom(channelName)
-    if gt:
-      for dploc in gt.getDeltas().getDeltaTab():
-        if dploc.noPt == refNo:
-          return dploc
-    return None
+    return next((dploc for dploc in gt.getDeltas().getDeltaTab() if dploc.noPt==refNo), None) if gt else None
+          
+
+  def deleteGTReference(self, grp, gt):
+    #  FIX 2019-05-05 : Some (old) character may have not group
+    if grp!=None:
+      # grp._lstAttr[:] = [ po for po in grp.getLstAttr() if (po.getPoserType()!=PoserToken.E_parmNode) or (gt.getName()==po.getValue()) ]
+      # grp.getLstAttr().remove(po)
+      for po in grp.getLstAttr():
+        if po.getPoserType()==PoserToken.E_parmNode:
+          if gt.getName()==po.getValue():
+            grp.getLstAttr().remove(po)
+            
+        elif po.getPoserType()==PoserToken.E_groupNode:
+          #  Recursive, because a group may contain another group
+          self.deleteGTReference(po, gt)
 
   # 
   # Delete the given channel.
   # 
   # @param gt
   def deleteChannel(self, gt):
-    self._channels.deleteChannel(gt)
+    del self._channels[gt.getName().lower()]
+    self.deleteGTReference(self._groups, gt)
 
   # 
   # Compute the first level list of channels that point to the given channel argument.
   # @param channel Reference channel
-  # @param lstIn  List to be extended and returned. Created if null. 
-  #                Does not contain duplicate channel.
-  def getChannelDescendant(self, channel, lstIn):
-    implst = lstIn if lstIn else [ ]
+  # @param setIn  Set to be extended and returned. Created if null. 
+  #               Does not contain duplicate channel.
+  def getChannelDescendant(self, channel, setIn):
+    implst = setIn if setIn else set()
     cn = channel.getName()
     gn = channel.getPoserMeshedObject().getName()
-    for gt in self.getChannels().getLstAttr():
+    for gt in self.getChannels():
       for vop in gt.getVOD():
-        if (gn==vop.getGroupName()) and (cn == vop.getChannelName()) and (not gt in implst):
-          implst.append(gt)
+        if (gn==vop.getGroupName()) and (cn==vop.getChannelName()) and (not gt in implst):
+          implst.add(gt)
           break
 
     return implst
+  
+  def getWeldParent(self):
+    return self.getFigure().getWeldParent(self.getName())
+
 
   # Test if the deltas are consistent with the geometry.
   # @param gt
@@ -645,7 +694,7 @@ class PoserMeshedObject(StructuredAttribut):
         else:
           res = C_BAD_DELTACOUNT
       else:
-            res = C_ERROR
+        res = C_ERROR
     else:
       res = C_NODELTA
     return res
@@ -656,7 +705,7 @@ class PoserMeshedObject(StructuredAttribut):
   def changeReference(self, oldPartName, newPartName, changeRefToBODY=True):
     self._parent = TAS(self._parent, oldPartName, newPartName)
     self._conformingTarget = TAS(self._conformingTarget, oldPartName, newPartName)
-    for gt in self._channels.getLstAttr():
+    for gt in self.getChannels():
       gt.changeReference(oldPartName, newPartName, changeRefToBODY=changeRefToBODY)
 
   def setDirect(self, tokenID, val):
@@ -700,15 +749,14 @@ class PoserMeshedObject(StructuredAttribut):
     elif tokenID==PoserToken.E_objFileGeom:
       self._geomType = PoserConst.GT_LOCAL_OBJFILE
 
-      tmps1 = val.strip()
       bci1 = 0
 
-      if tmps1.startswith(PoserConst.S_LOCAL_OBJFILE_PFX):
+      if val.startswith(PoserConst.S_LOCAL_OBJFILE_PFX):
         bci1 = len(PoserConst.S_LOCAL_OBJFILE_PFX) -1
       else:
-        bci1 = tmps1.rfind(' ')
+        bci1 = val.rfind(' ')
 
-      self._geomFileName = tmps1[(bci1 + 1):]
+      self._geomFileName = val[(bci1 + 1):]
 
     elif tokenID==PoserToken.E_endPoint:
       self._endPoint = Vector3d.parseVector3d(val)
@@ -734,9 +782,102 @@ class PoserMeshedObject(StructuredAttribut):
     else:
       logging.warning("Unexpected Direct: (%s) %s", tokenID.token, val)
 
+  
+  # Overload add Attribut
+  def addAttribut(self, po):
+    if isinstance(po, GenericTransform):
+      try:
+        self._channels[po.getName().lower()]
+        logging.warning("Duplicate Channel Definition [%s] in [%s] (overloaded)", po.getName(), self.getName())
+      except KeyError:
+        pass
+
+      self._channels[po.getName().lower()] = po
+      po.setPoserMeshedObject(self)
+      
+    elif po.getPoserType()==PoserToken.E_geomCustom:
+      self._geomCustom = po
+      self._geomType = PoserConst.GT_INTERNAL
+
+    elif po.getPoserType()==PoserToken.E_storageOffset:
+      # FIX20090607 : Storage Offset seen : shall not be kept in attribute list.
+      return
+
+    elif po.getPoserType()==PoserToken.E_alternateGeom:
+      self._lstAltGeom.append(po)
+      po.setNo(len(self._lstAltGeom))
+
+    elif po.getPoserType()==PoserToken.E_weightMap:
+      self._lstWMap.append(po)
+
+    elif po.getPoserType()==PoserToken.E_groups:
+      self._groups = po
+      
+    elif po.getPoserType()==PoserToken.E_readScript:
+      #  TODO : Read Channel Definition Script
+      logging.error("No Management rule for %s [%s]", po.getName(), po.getPoserType().token)
+    else:
+      super(PoserMeshedObject, self).addAttribut(po)
+
+  # Specific reader to take into account Channels class removal
+  #
+  def read(self, st):
+    inchannels = False
+
+    while True:
+      code,cn,rw = st.getLine()
+      if code==PoserFileParser.TT_WORD: # Other keyword found
+        try:
+          vc = Lang[cn]
+          #  Known word
+          if vc.isStructured:
+            if vc==PoserToken.E_channels:
+              # Read the Openning bracket and jump to next line
+              ncode,lw,rw = st.getLine()
+              if ncode==PoserFileParser.TT_LEFTBACKET:
+                inchannels = True
+                continue
+              else: # Syntax error
+                logging.warning("Line[%s] - read %s instead of '{'", st.lineno(), lw)
+                raise ParsingErrorException()
+
+            sta = create(vc, rw)
+            self.addAttribut(sta)
+            #  Read the opening bracket
+            code,cn,rw = st.getLine()
+            if code==PoserFileParser.TT_LEFTBACKET:
+              sta.read(st)
+            else:
+              if sta.isAmbi():
+                st.pushBack()
+              else:
+                logging.warning("Line[%s]:  '{' is missing for %s", st.lineno(), cn)
+                raise ParsingErrorException()
+          else:
+            if vc.isDirect:
+              self.setDirect(vc, rw)
+            else:
+              sa = create(vc, cn)
+              #  Read before add
+              sa.read(st, rw)
+              self.addAttribut(sa)
+        except KeyError:  #  Mot inconnu
+          logging.info("Line[%s] - Unknown word:%s", st.lineno(), cn)
+
+      elif code==PoserFileParser.TT_RIGHTBRACKET:
+        if inchannels:
+          inchannels = False
+        else:
+          break
+      elif code==PoserFileParser.TT_EOF:
+        break
+      else:
+        logging.warning("Line[%s] - Not Accepted :%s", st.lineno(), cn)
+        raise ParsingErrorException()
+
+    # End of read
+
   # Write the definition part of the meshed object.
-  # @param fw
-  # @param pfx
   def writeDef(self, fw, pfx):
     fw.write(pfx + self.getPoserType().token + " " + self.getName() + '\n')
     nextPfx = pfx + "\t"
@@ -756,19 +897,15 @@ class PoserMeshedObject(StructuredAttribut):
     fw.write(nextPfx + "}\n")
 
   # Write the Data part of the meshed object.
-  # @param fw
-  # @param pfx
   def writeData(self, fw, pfx):
     fw.write(pfx + self.getPoserType().token + " " + self.getName()+'\n')
     nextPfx = pfx + "\t"
     fw.write(nextPfx + "{\n")
     fw.write(nextPfx + "name " + self._pmName+'\n')
     fw.write(nextPfx + ("on" if self._visible else "off")+'\n')
-    #fw.write(nextPfx + "bend " + (1 if self._bend else 0)+'\n')
     WBoolLine(fw, nextPfx, "bend", self._bend)
     
     if self.getPoserFile().isVersion(PoserConst.POSER_V9):
-        #fw.write(nextPfx + "" + (1 if self._animatableOrigin else 0)+'\n')
         WBoolLine(fw, nextPfx, "animatableOrigin", self._animatableOrigin)
 
     WBoolLine(fw, nextPfx, "dynamicsLock", self._dynamicsLock)
@@ -796,7 +933,17 @@ class PoserMeshedObject(StructuredAttribut):
     for po in self._lstWMap:
       po.write(fw, nextPfx)
     
-    self._channels.write(fw, nextPfx)
+    # Channels { } part
+    # self._channels.write(fw, nextPfx)
+    fw.write(nextPfx + 'channels\n')
+    chanPfx = nextPfx + "\t"
+    fw.write(chanPfx + "{\n")
+    if self._groups:
+      self._groups.write(fw, chanPfx)
+    for po in self.getChannels():
+      po.write(fw, chanPfx)
+    fw.write(chanPfx + "}\n")
+
     if self._endPoint:
         fw.write(nextPfx + "endPoint " + self._endPoint.poserPrint()+'\n')
     if self._origin:
@@ -845,6 +992,16 @@ class Light(PoserMeshedObject):
   def getLightType(self):
     sa = self.findAttribut(PoserToken.E_lightType)
     return sa.getIntegerValue()
+  
+  def addAttribut(self, po):
+    if po.getPoserType()==PoserToken.E_depthMapSize:
+      prev = self.getGenericTransform(PoserToken.E_depthMapSize)
+      if prev:
+        # The second defintion should be the Simple attribut
+        self.getLstAttr().append(po)
+        return         
+      
+    PoserMeshedObject.addAttribut(self, po)
 
 # 
 #  Describes a Poser "actor" object.
@@ -878,12 +1035,6 @@ class PoserActor(PoserMeshedObject):
     self._castsShadow = False
     self._includeInDepthCue = False
 
-  def getWeldParent(self):
-    for po in self.getFigure().getFigureDesc().getLstAttr():
-      if (po.getPoserType() == PoserToken.E_weld) and (po.getChildName() == self.getName()):
-        return po.getParentName()
-    return None
-
 # 
 # Represents a Poser "prop" object
 #  
@@ -891,6 +1042,15 @@ class PoserProp(PoserMeshedObject):
   def __init__(self):
     super(PoserProp, self).__init__()
 
+  # Props may have a figure index
+  def setName(self, n):
+    super(PoserProp, self).setName(n)
+    self.idx = index(n)
+
+  def getIndex(self):
+    ''' Return the figure index or 0 if the prop is outside any figure ''' 
+    return self.idx
+
   def createChannelMorphList(self, stm):
     lstCS = ChannelMorphStatusList()
 
@@ -936,6 +1096,24 @@ class CurveProp(PoserProp):
     super(CurveProp, self).__init__()
 
 # 
+# 
+# Represents a Poser "Wave" object
+#  
+class WaveDeformerProp(PoserProp):
+  def __init__(self):
+    super(WaveDeformerProp, self).__init__()
+
+
+# 
+# 
+# Represents a Poser "Wind Field" object
+#  
+class ConeForceFieldProp(PoserProp):
+  def __init__(self):
+    super(ConeForceFieldProp, self).__init__()
+
+
+
 # Just to avoid interactions with actors or props that have geoms
 #  
 class ControlProp(PoserProp):
