@@ -4,9 +4,7 @@ Created on 3 oct. 2004
 Ported in Python in May 2020
 '''
 import sys
-import os 
 import logging
-import json
 import copy
 from collections import namedtuple
 
@@ -18,7 +16,7 @@ from pypos3d.wftk.WFBasic import Vector3d, Point3d, BoundingBox, BoundingSphere,
 from pypos3d.wftk.Repere import Repere
 from pypos3d.wftk.PoserFileParser import PoserFileParser, ParsingErrorException
 from pypos3d.pftk.PoserBasic import PoserConst, PoserObject, PToken, PoserToken, Lang, TAS, index, create, WBoolLine, RemoveQuotes, nodeNameNo
-from pypos3d.pftk.SimpleAttribut import KSA, AddChildSA, ValueOpDelta, OffSA
+from pypos3d.pftk.SimpleAttribut import KSA, ValueOpDelta, OffSA
 
  
 PtMapping = namedtuple('PtMapping', ['srcNo', 'refNo', 'refGeom', 'dist' ])
@@ -39,62 +37,43 @@ class StructuredAttribut(PoserObject):
   # and the opening bracket are supposed to be consumed by the caller.
   # Because it is the mean it uses to recognize a structured attribute.
   #    
-  def read(self, st):
-    nextLine = True
-    st.getToken()
-    while st.ttype != PoserFileParser.TT_EOF:
-      if st.ttype == PoserFileParser.TT_EOL:
-        st.getToken()
+  def read(self, st):    
+    code,cn,rw = st.getLine()
+    
+    while (code!=PoserFileParser.TT_EOF) and (code!=PoserFileParser.TT_RIGHTBRACKET):
       if st.ttype == PoserFileParser.TT_WORD:
-        if st.isRightBracket():
-          #  End of current structured attribute
-          break
-
-        cn = st.sval
         try:
-          vc = Lang[st.sval]
+          vc = Lang[cn]
           #  Known word
           if vc.isStructured:
-            nom = st.getToNextLine()
-            if st.ttype == PoserFileParser.TT_EOL:
-              nom = ""
-
-            sta = create(vc, nom)
+            sta = create(vc, rw)
 
             self.addAttribut(sta)
 
             #  Read the opening bracket
-            st.getToken()
-            if st.isLeftBracket():
+            code,cn,rw = st.getLine()
+            if code==PoserFileParser.TT_LEFTBACKET:
               sta.read(st)
             else:
               if sta.isAmbi():
-                #  Skip to next line
                 st.pushBack()
-                nextLine = False
               else:
                 logging.warning( "Line[%s]: '{' is missing for %s", st.lineno, cn)
           else:
             if vc.isDirect:
-              st.getToNextLine()
-              self.setDirect(vc, st.sval)
+              self.setDirect(vc, rw)
             else:
-              sa = create(vc, st.sval)
-              #  Try : 17AOU2007 : Read before add
-              sa.read(st)
+              sa = create(vc, cn)
+              #  Read before add
+              sa.read(st, rw)
               self.addAttribut(sa)
-            nextLine = False
 
         except KeyError: #  Mot inconnu
           logging.info("Line[%s] - Unknown word:%s", st.lineno(), cn)
 
-      if nextLine:
-        st.skipToNextLine()
-      else:
-        nextLine = True
-      #  Get next token
-      st.getToken()
-      # End of While
+      #  Get next line
+      code,cn,rw = st.getLine()
+      # Wend
 
   def isAmbi(self): return self.ambiguious
 
@@ -176,60 +155,41 @@ class VertsGroup(StructuredAttribut):
 
     fin = False
     while not fin:
-      st.getToken()
-      if st.ttype == PoserFileParser.TT_EOF:
-        fin = True
-        continue 
-
-      if st.sval == None:
-        continue 
-
-      if st.ttype == PoserFileParser.TT_WORD:
-        if st.sval == "v":
-          st.getNumber()
-          v = int(st.nval)
-          st.skipToNextLine()
-          self.vertTab.append(v)
+      code,cn,rw = st.getLine()
+      if code == PoserFileParser.TT_WORD:
+        if cn == "v":          
+          self.vertTab.append(int(rw))
           continue 
 
-        if st.isRightBracket():
-          fin = True
-        else:
-          #  Other keyword found
-          cn = st.sval
-          try:
-            vc = Lang[st.sval]
-            #  Known word
-            if vc.isStructured:
-              st.getToNextLine()
-              if st.ttype == PoserFileParser.TT_EOL:
-                nom = ""
-
-              # sta = vc.create(nom)
-              sta = create(vc, nom)
-              self.addAttribut(sta)
-              #  Read the opening bracket
-              st.getToken()
-              if st.isLeftBracket():
-                sta.read(st)
-              else:
-                if sta.isAmbi():
-                  #  Skip to next line
-                  st.pushBack()
-                else:
-                  logging.warning("Line[%s]:  '{' is missing for %s", st.lineno(), cn)
+        #  Other keyword found
+        try:
+          vc = Lang[cn]
+          #  Known word
+          if vc.isStructured:
+            sta = create(vc, rw)
+            self.addAttribut(sta)
+            #  Read the opening bracket
+            code,cn,rw = st.getLine()
+            if code==PoserFileParser.TT_LEFTBACKET:
+              sta.read(st)
             else:
-              #sa = vc.createSimple(st.sval)
-              sa = create(vc, st.sval)
-              #  Try : 17AOU2007 : Read before add
-              sa.read(st)
-              self.addAttribut(sa)
+              if sta.isAmbi():
+                st.pushBack()
+              else:
+                logging.warning("Line[%s]:  '{' is missing for %s", st.lineno(), cn)
+          else:
+            sa = create(vc, cn)
+            #  Read before add
+            sa.read(st, rw)
+            self.addAttribut(sa)
 
-          except:
-            #  Mot inconnu
-            logging.info("Line[%s] - Unknown word:%s",st.lineno() ,  cn)
+        except:
+          #  Mot inconnu
+          logging.info("Line[%s] - Unknown word:%s",st.lineno(),  cn)
+      elif (code==PoserFileParser.TT_EOF) or (code==PoserFileParser.TT_RIGHTBRACKET): 
+        break
       else:
-        logging.warning("Line[%s] - Not Accepted :%s", st.lineno() , st.sval)
+        logging.warning("Line[%s] - Not Accepted :%s", st.lineno(), cn)
         raise ParsingErrorException()
 
     # End of read
@@ -306,17 +266,10 @@ class NodeInput(StructuredAttribut):
   def read(self, st):
        
     while True:
-      st.getToken()
-      if st.ttype==PoserFileParser.TT_EOF or st.isRightBracket():
-        break
-
-      if not st.sval:
-        continue
-
-      if st.ttype == PoserFileParser.TT_WORD:
+      code,cn,val = st.getLine()
+      if code==PoserFileParser.TT_WORD:
         try:
-          tokenID = Lang[st.sval]        
-          val = st.getToNextLine()
+          tokenID = Lang[cn]        
         
           if tokenID==PoserToken.E_name:
             self._nodeName = RemoveQuotes(val)
@@ -334,15 +287,14 @@ class NodeInput(StructuredAttribut):
             val = val.strip()
             self._file = None if val==PoserConst.C_NO_MAPG or val==PoserConst.C_NO_MAP else val
           else:
-            logging.warning("Unexpected Direct:%s (%s)",tokenID.token, val)
+            logging.warning("Unexpected Direct in NodeInput:%s (%s)",tokenID.token, val)
         
         except KeyError: #  Mot inconnu
-          logging.info("Line[%s] - Unknown word:%s", st.lineno(), st.sval)
-
-        if st.isRightBracket():
-          break
+          logging.info("Line[%s] - Unknown word:%s", st.lineno(), cn)
+      elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==PoserFileParser.TT_EOF): 
+        break
       else:
-        logging.warning("L[%d] - Not Accepted :%s", st.lineno(), st.sval)
+        logging.warning("L[%d] - Not Accepted :%s", st.lineno(), cn)
         raise ParsingErrorException()
 
 
@@ -401,7 +353,7 @@ class Node(StructuredAttribut):
     if tokenID == PoserToken.E_name:
       self._nodeName = RemoveQuotes(val)
     else:
-      logging.warning("Unexpected Direct:%s (%s)",tokenID.token, val)
+      logging.warning("Unexpected Direct in Node:%s (%s)",tokenID.token, val)
   
   def addAttribut(self, po):
     if po.getPoserType()==PoserToken.E_pos:
@@ -582,7 +534,7 @@ class InkyChain(StructuredAttribut):
     elif tokenID==PoserToken.E_off:
       self._active = False
     else:
-      logging.warning("Unexpected Direct:%s (%s)", tokenID.token, val)
+      logging.warning("Unexpected Direct in InkyChain:%s (%s)", tokenID.token, val)
 
   # 
   #    * @see pftk.poser.kern.StructuredAttribut#write(java.io.PrintWriter,
@@ -626,48 +578,46 @@ class Keys(StructuredAttribut):
     self._hshKeys = { frameNo:KSA(src=ksa) for frameNo,ksa in src._hshKeys.items() } if src else { } # Replaced by a dict frameNo:KSA
 
   def read(self, st):
-    fin = False
     curK = None
-    while not fin:
-      st.getToken()
-      if st.ttype == PoserFileParser.TT_EOF:
-        fin = True
-        continue 
-
-      if st.sval == None:
-        continue 
-
-      if st.ttype == PoserFileParser.TT_WORD:
+    while True:
+      code,cn,rw = st.getLine()
+      if code==PoserFileParser.TT_WORD:
         try:
-          vc = Lang[st.sval]
+          vc = Lang[cn]
 
-          if vc==PoserToken.E_static:
-            self._static = st.getBoolToNextLine()
-            
-          elif vc==PoserToken.E_k:
-            curK = KSA(int(st.getNumber()), st.getNumber())
+          if vc==PoserToken.E_k:
+            tv = rw.split()
+            curK = KSA(int(tv[0]), float(tv[1]))
             self._hshKeys[curK._noFrame] = curK
-
-          elif vc==PoserToken.E_sl:
-            curK.setSl(st.getBoolToNextLine())
+            # Read 3 lines after (if any)
+            c0, csl, cval = st.getLine()
             
-          elif (vc==PoserToken.E_spl) or (vc==PoserToken.E_lin) or (vc==PoserToken.E_con):
-            curK.setCurveType(vc)
+            if c0==PoserFileParser.TT_RIGHTBRACKET: 
+              break
+            
+            c1, cty, *_ = st.getLine()
+            c2, ccnx, *_ = st.getLine()
 
-          elif (vc==PoserToken.E_sm) or (vc==PoserToken.E_br):
-            curK.setCurveCnx(vc)
+            # elif vc==PoserToken.E_sl:
+            curK.setSl(cval[0]=='1')
+            
+            # elif (vc==PoserToken.E_spl) or (vc==PoserToken.E_lin) or (vc==PoserToken.E_con):
+            curK.setCurveType(Lang[cty])
+
+            #elif (vc==PoserToken.E_sm) or (vc==PoserToken.E_br):
+            curK.setCurveCnx(Lang[ccnx])
+            
+          elif vc==PoserToken.E_static:
+            self._static = (rw[0]=='1')
           else:
-            logging.warning("L[%d] - Not Accepted :%s", st.lineno(), st.sval)
+            logging.warning("L[%d] - Not Accepted :%s", st.lineno(), cn)
             
-          continue 
         except KeyError:
           pass
-
-        if st.isRightBracket():
-          fin = True
-
+      elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==PoserFileParser.TT_EOF): 
+        break
       else:
-        logging.warning("L[%d] - Not Accepted :%s", st.lineno(), st.sval)
+        logging.warning("L[%d] - Not Accepted :%s", st.lineno(), cn)
         raise ParsingErrorException()
 
 
@@ -753,7 +703,7 @@ class DocDescription(StructuredAttribut):
     elif tokenID==PoserToken.E_displayMode:
       self._displayMode = val
     else:
-      logging.warning("Unexpected Direct:%s (%s)", tokenID.token, val)
+      logging.warning("Unexpected Direct in DocDescription:%s (%s)", tokenID.token, val)
 
   #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
   def write(self, fw, pfx):
@@ -774,108 +724,9 @@ class DocDescription(StructuredAttribut):
 
     fw.write(nextPfx + "}\n")
 
-
-
-
-
-
-
-# 
 # Description part of a figure.
-#  
-class FigureDescription(StructuredAttribut):
-
-  def __init__(self):
-    super(FigureDescription, self).__init__()
-    self._displayMode = None
-    self._locked = False
-    self._root = None # Root Actor Name
-
-  # 
-  #    * Change the name of referenced part (actor, prop, hairProp, controlProp)  
-  #    * @param oldPartName
-  #    * @param newPartName
-  #    
-  def changeReference(self, oldPartName, newPartName):
-    #  Change Attributes : addChild and weld
-    for po in self.getLstAttr():
-      if (po.getPoserType() == PoserToken.E_root) or (po.getPoserType() == PoserToken.E_defaultPick):
-        po.TAS(oldPartName, newPartName)
-      else:
-        if isinstance(po, AddChildSA):
-          po.changeReference(oldPartName, newPartName)
-        else:
-          if po.getPoserType() == PoserToken.E_inkyChain:
-            for po2 in po.getLstAttr():
-              if (po2.getPoserType() == PoserToken.E_goal) or (po2.getPoserType() == PoserToken.E_addLink):
-                po2.TAS(oldPartName, newPartName)
-
-  # 
-  #  Add the list of actors as starting from pa, following pa's figure child and weld trees
-  #  * @param parent
-  #  * @param pa
-  #    
-  def addActor(self, parent, ldesc, weldActors=True):
-    seen = False
-    last = -1
-
-    for i,po in enumerate(self.getLstAttr()):
-      if po.getPoserType() == PoserToken.E_addChild:
-        seen = True
-      elif seen: #  Add before "i"
-        last = i
-        for pfils in ldesc:
-          ac = AddChildSA(PoserToken.E_addChild, pfils.getName(), pfils.getParent())
-          self.getLstAttr().insert(last, ac)
-          last += 1
-        break
-
-    if weldActors:
-      seen = False
-      for i,po in enumerate(self.getLstAttr()):
-        #po = self.getLstAttr[i]
-        if po.getPoserType() == PoserToken.E_weld:
-          seen = True
-        elif seen: #  Add before "i"
-          last = i
-          for pfils in ldesc:
-            gwp = pfils.getWeldParent()
-            if gwp:
-              ac = AddChildSA(PoserToken.E_weld, pfils.getName(), gwp)
-              self.getLstAttr().insert(last, ac)
-              last += 1
-          break
+#  class FigureDescription(StructuredAttribut): --> inlined in Figure
         
-  def getRoot(self): return self._root
-
-  def setDirect(self, tokenID, val):
-    if tokenID==PoserToken.E_name:
-      self.setName(val)
-    elif tokenID==PoserToken.E_locked:
-      self._locked = (val[0] == "1")
-    elif tokenID==PoserToken.E_displayMode:
-      self._displayMode = val
-    elif tokenID==PoserToken.E_root:
-      self._root = val
-    else:
-      logging.warning("Unexpected Direct:%s (%s)", tokenID.token, val)
-
-  #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
-  def write(self, fw, pfx):
-    fw.write(pfx + self.getPoserType().token+'\n')
-    nextPfx = pfx + "\t"
-    fw.write(nextPfx + "{\n")
-    fw.write(nextPfx + "name " + self.getName() + '\n')
-    fw.write(nextPfx + "root " + self.getRoot() + '\n')
-    
-    for po in self._lstAttr:
-      po.write(fw, nextPfx)
-    if self._displayMode:
-      fw.write(nextPfx + "displayMode " + self._displayMode + '\n')
-
-    WBoolLine(fw, nextPfx, "locked", self._locked )
-
-    fw.write(nextPfx + "}\n")
 
 class AlternateGeom(StructuredAttribut):
   def __init__(self):
@@ -890,12 +741,11 @@ class AlternateGeom(StructuredAttribut):
   def setDirect(self, tokenID, val):
     if tokenID==PoserToken.E_name:
       self._printName = val
-    elif tokenID==PoserToken.E_objFile:
-      tmps1 = val.strip()
-      bci1 = tmps1.rfind(' ')
-      self._geomFileName = tmps1[(bci1 + 1):]
+    elif tokenID==PoserToken.E_objFile:      
+      bci1 = val.rfind(' ')
+      self._geomFileName = val[(bci1 + 1):]
     else:
-      logging.warning("Unexpected Direct:%s (%s)", tokenID.token, val)
+      logging.warning("Unexpected Direct in AlternateGeom:%s (%s)", tokenID.token, val)
 
   #  @see pftk.poser.kern.StructuredAttribut#write(java.io.PrintWriter,
   def write(self, fw, pfx):
@@ -962,30 +812,22 @@ class Deltas(StructuredAttribut):
     self.setPoserType(PoserToken.E_deltas)
 
     #In python deltaTab replaced by a dict { noPt:DeltaPoint() }
-    self.deltaSet = { dp.noPt:DeltaPoint(dp.noPt, dp.x, dp.y, dp.z) for dp in src.deltaSet.items() } if src else { }
+    self.deltaSet = { no:DeltaPoint(dp.noPt, dp.x, dp.y, dp.z) for no,dp in src.deltaSet.items() } if src else { }
 
   #    * @see deyme.v3d.poser.PoserObject#read(deyme.v3d.poser.PoserFileParser)
   def read(self, st):
     self.deltaSet = { }
-    fin = False
-    while not fin:
-      st.getToken()
-      if st.ttype == PoserFileParser.TT_EOF:
+
+    while True:
+      code,cn,rw = st.getLine()
+      if (code==PoserFileParser.TT_WORD) and (cn=="d"):
+        tv = rw.split()
+        p = DeltaPoint(int(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
+        self.deltaSet[p.noPt] = p
+      elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==PoserFileParser.TT_EOF): 
         break
-
-      if not st.sval:
-        continue
-
-      if st.ttype == PoserFileParser.TT_WORD:
-        if st.sval == "d":
-          p = DeltaPoint(int(st.getNumber()), st.getNumber(), st.getNumber(), st.getNumber())
-          st.skipToNextLine()
-          self.deltaSet[p.noPt] = p
-          continue 
-        if st.isRightBracket():
-          fin = True
       else:
-        logging.warning("L[%d] - Not Accepted :%s", st.lineno(), st.sval)
+        logging.warning("L[%d] - Not Accepted :%s", st.lineno(), cn)
         raise ParsingErrorException()
 
   # @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
@@ -1145,6 +987,9 @@ class Deltas(StructuredAttribut):
 #  This method should actually take a ValueOpDelta[] because a single
 #  transform can have many valueOpDelta fields!!!
 class GenericTransform(StructuredAttribut):
+  # Managed operators
+  OPS = (PoserToken.E_valueOpDeltaAdd, PoserToken.E_valueOpPlus, PoserToken.E_valueOpMinus, PoserToken.E_valueOpTimes, PoserToken.E_valueOpDivideBy, PoserToken.E_valueOpDivideInto, PoserToken.E_valueOpKey)
+
   def __init__(self, poserType=None, channelName=None, vodFigure=None, vodActor=None, vodChannel=None):
     super(GenericTransform, self).__init__()
 
@@ -1185,7 +1030,7 @@ class GenericTransform(StructuredAttribut):
       #  Create the deltas table
       if poserType != PoserToken.E_valueParm:
         self._numbDeltas = 0
-        self._dlt = Deltas()
+        #self._dlt = Deltas()
 
         if vodActor or vodChannel:
           if vodFigure:
@@ -1197,6 +1042,14 @@ class GenericTransform(StructuredAttribut):
             vod = ValueOpDelta(PoserToken.E_valueOpDeltaAdd, vodFigure, vodActor, vodChannel, 1.0)
             self.addVOD(vod)
 
+
+
+  # Overload setPoserType to create Deltas
+  def setPoserType(self, poserType):
+    StructuredAttribut.setPoserType(self, poserType)
+    if not self._dlt and (poserType==PoserToken.E_targetGeom):
+      self._dlt = Deltas()
+
   #
   # Copy all data from src, except the _pmo if parentPO is set
   #
@@ -1234,7 +1087,8 @@ class GenericTransform(StructuredAttribut):
   def getKeysFactor(self, frameNo):
     return self._keys.getKeyFactor(frameNo) if self._keys else 0.0
 
-  def getDeltas(self): return self._dlt
+  def getDeltas(self):
+    return self._dlt
 
   def ishasDeltas(self): return self._dlt and (len(self._dlt.deltaSet)>0)
 
@@ -1255,12 +1109,11 @@ class GenericTransform(StructuredAttribut):
   #    
   def optimizeDeltas(self, refNorm):
     dl = self.getDeltas()
-    #t = dl.getDeltaTab()
     oldLength = len(dl.deltaSet)
-    optSet = { t.noPt:t for t in dl.deltaSet if t.norm() > refNorm }
+    optSet = { no:t for no,t in dl.deltaSet.items() if t.norme() > refNorm }
 
     # optTabFinal = Arrays.copyOf(optTab, nbpt)
-    dl.setDeltaTab(optSet)
+    dl.deltaSet = optSet
     indexes = len(optSet)
     logging.info("Channel[%s] reduced from %d to %d", self.getName(), oldLength, indexes )
     return indexes 
@@ -1272,11 +1125,8 @@ class GenericTransform(StructuredAttribut):
     self._lstVOD.append(vop)
 
   def findValueOp(self, vodType:'ValueOpDelta', targetName:'str'=None):
-    for vop in self._lstVOD:
-      if (vop.getPoserType()==vodType) and ((targetName is None) or (vop.getChannelName()==targetName)):
-        return vop
-       
-    return None
+    return next((vop for vop in self._lstVOD \
+                if (vop.getPoserType()==vodType) and ((targetName is None) or (vop.getChannelName()==targetName))), None)
 
   # 
   #    * Delete any reference to the given GenericTransform
@@ -1286,7 +1136,6 @@ class GenericTransform(StructuredAttribut):
     chName = gt.getName()
     grpName = gt.getPoserMeshedObject().getName()
     self._lstVOD[:] = [ vod for vod in self._lstVOD if (grpName!=vod.getGroupName()) or (chName!=vod.getChannelName()) ]
-    #self._vodTab = None
 
 
   # 
@@ -1296,7 +1145,6 @@ class GenericTransform(StructuredAttribut):
   def deleteFigureRef(self, fig):
     fi = fig.getBodyIndex()
     self._lstVOD[:] = [ vod for vod in self._lstVOD if index(vod.getGroupName())!=fi ]
-    #self._vodTab = None
 
 
   # 
@@ -1313,7 +1161,6 @@ class GenericTransform(StructuredAttribut):
       if po.getPoserType() == PoserToken.E_otherActor:
         po.TAS(oldPartName, newPartName)
 
-  OPS = (PoserToken.E_valueOpDeltaAdd, PoserToken.E_valueOpPlus, PoserToken.E_valueOpMinus, PoserToken.E_valueOpTimes, PoserToken.E_valueOpDivideBy, PoserToken.E_valueOpDivideInto, PoserToken.E_valueOpKey) 
 
   def addAttribut(self, po):
     if po.getPoserType() in GenericTransform.OPS:
@@ -1355,7 +1202,7 @@ class GenericTransform(StructuredAttribut):
     elif tokenID==PoserToken.E_staticValue:
       self.setStaticValue(float(val))
     else:
-      logging.warning("Unexpected Direct:%s (%s)", tokenID.token, val)
+      logging.warning("Unexpected Direct in GenericTranform:%s (%s)", tokenID.token, val)
 
   #    * @see pftk.poser.kern.StructuredAttribut#write(java.io.PrintWriter,
   def write(self, fw, pfx):
@@ -1398,6 +1245,9 @@ class GenericTransform(StructuredAttribut):
 
   def getPrintName(self): return self._transfName
 
+  def getQualifiedName(self):
+    return (self._pmo.getName() if self._pmo else '?' )+ '.' + self.getName()
+
   def setPrintName(self, n):
     self._transfName = n
 
@@ -1473,10 +1323,7 @@ class GenericTransform(StructuredAttribut):
   # @return    true when the set of target morph is empty AND the attrName starts with "PBM" OR when the set of target contains the given attrName
   @classmethod
   def concerned(cls, attrName, setTargetMorph):
-    if setTargetMorph == None:
-      return attrName.startswith("PBM")
-    else:
-      return attrName in setTargetMorph
+    return attrName.startswith("PBM") if setTargetMorph==None else attrName in setTargetMorph 
 
 
 class Scale(GenericTransform):
@@ -1520,154 +1367,7 @@ class DepthMapSize(GenericTransform):
 
 # 
 # Represents the list of channels of a PoserMeshedObject.
-#  
-class Channels(StructuredAttribut):
-
-  def __init__(self):
-    super(Channels, self).__init__()
-    self.setPoserType(PoserToken.E_channels)
-    self.setName("")
-
-    #  PoserMeshedObject that contains the channel list 
-    self._pmo = None
-
-    #  Store the group presentation of channels 
-    self._groups = None
-
-  def setPoserMeshedObject(self, pmo):
-    self._pmo = pmo
-
-  def getPoserMeshedObject(self): return self._pmo
-
-  # 
-  #    * Delete the given channel.
-  #    * @param gt
-  #    
-  def deleteChannel(self, gt):
-    self.deleteAttribut(gt.getName())
-    self.deleteGTReference(self._groups, gt)
-
-  def deleteGTReference(self, grp, gt):
-    #  FIX 2019-05-05 : Some (old) character may have not group
-    if grp != None:
-      # grp._lstAttr[:] = [ po for po in grp.getLstAttr() if (po.getPoserType()!=PoserToken.E_parmNode) or (gt.getName()==po.getValue()) ]
-      # grp.getLstAttr().remove(po)
-
-      for po in grp.getLstAttr():
-        if po.getPoserType()==PoserToken.E_parmNode:
-          if gt.getName()==po.getValue():
-            grp.getLstAttr().remove(po)
-            
-        elif po.getPoserType()==PoserToken.E_groupNode:
-          #  Recursive, because a group may contain another group
-          self.deleteGTReference(po, gt)
-
-  # 
-  #    * Add the given channel.
-  #    * @param gt
-  #    
-  def addChannel(self, gt):
-    res = C_OK
-    po = self.findAttribut(gt.getPoserType(), gt.getName())
-    if not po:
-      self.getLstAttr().append(gt)
-    else:
-      logging.info("Channel[%s:%s] already exists", gt.getPoserType(), gt.getName()  )
-      res = C_FAIL
-    return res
-
-  # 
-  # Update or create a targetGeom and create a dependency to the body homonyme valueParm (if bodyLinked=true)
-  #
-  # @param channelName
-  #    *          TargetGeom name
-  # @param targetFigureName
-  #    *          Referenced Figure Name
-  # @param srcWG
-  #    *          Concerned WaveGeom
-  # @param lstNewDeltas
-  #    *          List of deltas
-  # @param bodyIdx
-  #    *          Body index of the referenced figure (silly)
-  #    
-  #     updateOrCreate(String channelName, String targetFigureName, WaveGeom3d srcWG, lstNewDeltas, bodyIdx, bodyLinked)
-  # def updateOrCreate_0(self, channelName, targetFigureName, srcWG, lstNewDeltas, bodyIdx):
-  def updateOrCreate(self, channelName, targetFigureName, srcWG, setNewDeltas, bodyIdx, bodyLinked=True):
-    gt = None
-    #  trouve -> "found"
-    trouve = False
-    for gt in self.getLstAttr():
-      trouve = (gt.getPoserType() == PoserToken.E_targetGeom) and (gt.getName() == channelName)
-      if trouve: break
-
-    if not trouve: #  Create the missing targetGeom
-      if bodyLinked:
-        gt = GenericTransform(PoserToken.E_targetGeom, channelName, targetFigureName, \
-                              None if not targetFigureName else "BODY:" + str(bodyIdx),\
-                              None if not targetFigureName else channelName)
-      else:
-        gt = GenericTransform(PoserToken.E_targetGeom, channelName)
-        #  Do not add a prefix to printable name (old P4 Stuff)
-        gt.setPrintName(channelName[3:] if (channelName.startswith("PBM") or channelName.startswith("FBM")) else channelName)
-        gt.addKeyFrame(0, 0.0)
-        #  Default keys are static=0
-        gt.setInterpStyleLocked(0)
-
-      self.getLstAttr().insert(0, gt)
-
-    gt.getDeltas().deltaSet = setNewDeltas # [ delta for delta in lstNewDeltas ] )
-    gt.setNumbDeltas(len(srcWG.coordList))
-    return gt
-
-  # 
-  # Update or create a valueParm and initializes it
-  # @param channelName ValueParm name
-  def updateOrCreateVP(self, channelName, minVal=0.0, maxVal=1.0, applyLimits=True, isHidden=False):
-    vp = None
-    #  trouve -> "found"
-    trouve = False
-    for vp in self.getLstAttr():
-      trouve = (vp.getPoserType()==PoserToken.E_valueParm) and (vp.getName()==channelName)
-      if trouve: break
-
-    if not trouve:
-      #  Create the missing valueParm
-      vp = GenericTransform(PoserToken.E_valueParm, channelName)
-      #  Do not add a prefix to printable name (old P4 Stuff)
-      vp.setPrintName(channelName[3:] if (channelName.startswith("PBM") or channelName.startswith("FBM")) else channelName)
-      vp.addKeyFrame(0, 0.0)
-      #  Default keys are static=0
-      vp.setInterpStyleLocked(0)
-
-      self.getLstAttr().insert(0, vp)
-
-    vp.setMin(minVal)
-    vp.setMax(maxVal)
-    vp.setForceLimits(1 if applyLimits else 0)
-    vp.setHidden(isHidden)
-    return vp
-
-  def addAttribut(self, po):
-    if po.getPoserType()==PoserToken.E_groups:
-      self._groups = po
-    elif po.getPoserType()==PoserToken.E_readScript:
-      #  TODO : Read Channel Definition Script
-      logging.error("No Management rule for %s [%s]", po.getName(), po.getPoserType().token)
-    else:
-      po.setPoserMeshedObject(self._pmo)
-      super(Channels, self).addAttribut(po)
-
-  def write(self, fw, pfx):
-    fw.write(pfx + self.getPoserType().token+'\n')
-    nextPfx = pfx + "\t"
-    fw.write(nextPfx + "{\n")
-    if self._groups:
-      self._groups.write(fw, nextPfx)
-    for po in self._lstAttr:
-      po.write(fw, nextPfx)
-    if len(self._lstAttr) == 0:
-      fw.write('\n')
-    fw.write(nextPfx + "}\n")
+# class Channels(StructuredAttribut):  --> Inlined in PoserMeshedObject
 
 # 
 #  * Used to create morph. 
@@ -1746,7 +1446,6 @@ class ChannelMorphStatus(object):
 # A list of ChannelMorphStatus
 #  
 class ChannelMorphStatusList(list):
-  CMSL_FILE_VERSION = 0xf000000016112007
   FILTER_NONE = 0x0000
   FILTER_NONNULLREFDELTA = 0x0001
   FILTER_REFPRESENCE = 0x0002
@@ -1808,36 +1507,36 @@ class ChannelMorphStatusList(list):
   # 
   #  Save the list of selected channels.
   #  @param fn  Full path name of the file to create.
-  #    
-  def saveSelected(self, fn):
-    try:
-      #  Write File Version
-      ois = os.open(fn, "w")
-
-      jsonD = { cs.channelName : cs for cs in self if cs.selected }
-
-      json.dump(jsonD, ois)
-
-      ois.close()
-      logging.info("List saved in %s (JSON)", fn)
-    except IOError as ioex:
-      ioex.printStackTrace()
-
-  @classmethod
-  def loadList(cls, fn):
-    l = ChannelMorphStatusList()
-    try:
-      ois = os.open(fn, "r")
-      data = json.load(ois)
-      for cs in data.values():
-        l.append(cs)
-
-      ois.close()
-    except IOError as ioex:
-      # ioex.printStackTrace()
-      print(str(ioex))
-      l = None
-    return l
+  #TODO: Useless and serialization pb due to the link to a Generictransform
+#   def saveSelected(self, fn):
+#     try:
+#       #  Write File Version
+#       ois = open(fn, "w")
+# 
+#       jsonD = { cs.channelName : cs.__dict__ for cs in self if cs.selected }
+# 
+#       json.dump(jsonD, ois)
+# 
+#       ois.close()
+#       logging.info("List saved in %s (JSON)", fn)
+#     except IOError as ioex:
+#       ioex.printStackTrace()
+# 
+#   @classmethod
+#   def loadList(cls, fn):
+#     l = ChannelMorphStatusList()
+#     try:
+#       ois = open(fn, "r")
+#       data = json.load(ois)
+#       for cs in data.values():
+#         l.append(cs)
+# 
+#       ois.close()
+#     except IOError as ioex:
+#       # ioex.printStackTrace()
+#       print(str(ioex))
+#       l = None
+#     return l
 
 
 # 
@@ -1858,7 +1557,7 @@ class FrontImage(StructuredAttribut):
     elif tokenID==PoserToken.E_flipped:
       self._flipped = val.startswith("1")
     else:
-      logging.warning("Unexpected Direct:%s (%s)", tokenID.token, val)
+      logging.warning("Unexpected Direct in FrontImage:%s (%s)", tokenID.token, val)
 
   #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
   def write(self, fw, pfx):
@@ -1891,34 +1590,20 @@ class WeightMap(StructuredAttribut):
   def read(self, st):
     self._vertTab = [ ]
 
-    fin = False
-    while not fin:
-      st.getToken()
-      if st.ttype == PoserFileParser.TT_EOF:
-        fin = True
-        continue 
-
-      if st.sval == None:
-        continue 
-
-      if st.ttype == PoserFileParser.TT_WORD:
-        if st.sval == 'numbVerts':          
-          self._numbVerts = int(st.getNumber())
-          st.skipToNextLine()
-          continue
-        
-        if st.sval == "v":
-          self._vertTab.append( ( int(st.getNumber()), st.getNumber() ) )
-          st.skipToNextLine()
-          continue 
-
-        if st.isRightBracket():
-          fin = True
-        else:
-          #  Mot illegal
-          logging.info("Line[%s] - Not accepted word:%s",st.lineno() ,  st.sval)
+    while True:
+      code,cn,rw = st.getLine()
+      if code== PoserFileParser.TT_WORD:
+        if cn=="v":
+          tv = rw.split()
+          self._vertTab.append( ( int(tv[0]), float(tv[1]) ) )
+        elif cn=='numbVerts':
+          self._numbVerts = int(rw)
+        else: #  Mot illegal
+          logging.info("Line[%s] - Not accepted word:%s",st.lineno(), cn)
+      elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==PoserFileParser.TT_EOF): 
+        break
       else:
-        logging.warning("Line[%s] - Not Accepted :%s", st.lineno() , st.sval)
+        logging.warning("Line[%s] - Not Accepted :%s", st.lineno(), cn)
         raise ParsingErrorException()
 
     # End of read
