@@ -45,14 +45,14 @@ def getOBJFile(filename):
   p = lfn.rfind('.obj')
 
   if os.path.isfile(filename):
-    fic = open(filename, 'rb') if p>0 else gzip.open(filename, 'rb') 
+    fic = open(filename, 'rt', errors='replace') if p>0 else gzip.open(filename, 'rt', errors='replace') 
   else:
     # Change for the other extension
     if p>0:
-      fic = gzip.open(filename[0:p] + '.obz', 'rb') 
+      fic = gzip.open(filename[0:p] + '.obz', 'rt', errors='replace') 
     else:
       p=lfn.rfind('.obz')
-      fic = open(filename[0:p] + '.obj', 'rb') 
+      fic = open(filename[0:p] + '.obj', 'rt', errors='replace') 
 
   return fic
 
