@@ -130,10 +130,13 @@ class WaveGeom(object):
     try:
       st = kwargs['st']
       self.readInternal(st)
-    except:
+    except KeyError:
       # When the WaveGeom is not created from a file 
       # a default material is required
       self.addMaterial("default") 
+      
+    #except Exception as e: # Should
+    #  logging.error("Error:%s", str(e))
 
   def setName(self, n):
     self._name = n
@@ -262,7 +265,8 @@ class WaveGeom(object):
     wg.groups = [ GeomGroup(src=ggd, inCoordList=wg.coordList) for ggd in self.groups ]
 
     # Link correctly groups
-    self.optimizeGroups()
+    # FIX a major error
+    wg.optimizeGroups()
 
     return wg
 
@@ -394,97 +398,85 @@ class WaveGeom(object):
 
     fin = False
     while not fin:
-      st.getToken()
-
-      if st.ttype == PoserFileParser.TT_EOF:
-        fin = True
-        continue
-
-      if st.sval == None:
-        continue
-
-      if st.ttype == PoserFileParser.TT_WORD:
-        if st.sval=="v":
+      code,lw,rw = st.getLine()
+      
+      if code == PoserFileParser.TT_WORD:
+        if lw=="v":
           self.coordList.append(st.readVertex())
           continue
 
-        if st.sval=="vt":
+        if lw=="vt":
           self.texList.append(st.readTexture())
           continue
 
-        if st.sval=="vn":
+        if lw=="vn":
           # Ignore normals
           self.normList.append(st.readNormal())
           continue
 
-        if st.sval=="f":
+        if lw=="f":
           self.readFace(st)
           continue
 
-        if st.sval=="l":
+        if lw=="l":
           self.readLine(st)
           continue
 
-        if st.sval=="numbVerts":
-          st.getToken()
-          numbVerts = int(st.nval)
-
-        if st.sval=="numbTVerts":
-          st.getToken()
-          numbTVerts = int(st.nval)
-
-        if st.sval=="numbTSets":
-          st.getToken()
-          numbTSets = int(st.nval)
-
-        if st.sval=="numbElems":
-          st.getToken()
-          numbElems = int(st.nval)
-
-        if st.sval=="numbSets":
-          st.getToken()
-          numbSets = int(st.nval)
-
-        if st.sval=="usemtl":
-          # FIX 20110314-ODY: Material names could contain white chars and others (like #)
-          st.getToNextLine()
-          self.addMaterial(st.sval)
-
-        if st.sval=="s":
-          st.skipToNextLine()
+        if lw=="numbVerts":
+          numbVerts = int(rw)
           continue
 
-        if st.sval=="g":
-          st.getToken() #  should read a NAME
-          self.curGroup = self.createGeomGroup(st.sval)
+        if lw=="numbTVerts":
+          numbTVerts = int(rw)
+          continue
+
+        if lw=="numbTSets":
+          numbTSets = int(rw)
+          continue
+
+        if lw=="numbElems":
+          numbElems = int(rw)
+          continue
+
+        if lw=="numbSets":
+          numbSets = int(rw)
+          continue
+
+        if lw=='usemtl':
+          # FIX 20110314-ODY: Material names could contain white chars and others (like #)
+          self.addMaterial(rw)
+          continue
+
+        if lw=="s":
+          continue
+
+        if lw=="g":
+          self.curGroup = self.createGeomGroup(rw)
           self.curGroup.setMaterialName(self.curMatIdx)
           continue
 
         # Record mtllibs that could be read and resolved later
-        if st.sval=="mtllib":
-          st.getToken() #  should read a FILENAME
-          if st.getFileNature() == OBJ_FNAT:
-            self._libMaterialName = st.sval
-            # log.info("mtllib " + st.sval + " stored")
-          else:
-            # log.info("mtllib " + st.sval + " ignored")
-            pass
+        if lw=='mtllib': # rw should read a FILENAME
+          if st.getFileNature()==OBJ_FNAT:
+            self._libMaterialName = rw
           continue
 
-        if (st.sval[0]=="g") and (len(st.sval) > 1):
+        if (lw[0]=="g") and (len(st.sval) > 1):
           # Strange case where there's no space after 'g' (OBJ regular ??)
-          self.curGroup = self.createGeomGroup(st.sval[1:])
+          self.curGroup = self.createGeomGroup(lw[1:])
           self.curGroup.setMaterialName(self.curMatIdx)
           continue
 
-        if st.isRightBracket():
-          fin = True
+      elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==code==PoserFileParser.TT_EOF):
+        fin = True
       else:
-        # log.warning("L[" + st.lineno() + "] - Not Accepted :" + st.sval)
-        raise Exception("L[%d] - Not Accepted : %s", st.lineno(), st.sval)
+        # log.warning("L[" + st.lineno() + "] - Not Accepted :" + lw)
+        raise Exception("L[%d] - Not Accepted : %s", st.lineno(), lw)
 
-    if (st.getFileNature() != OBJ_FNAT) and not self.groups:
-      self.groups.append(self.curGroup)
+    # Ensure that the default group is kept (may append)
+    if ((st.getFileNature() != OBJ_FNAT) and not self.groups) or \
+        (not self.curGroup in self.groups):
+      self.groups.append(self.curGroup)      
 
     # Close Groups
     self.optimizeGroups()
@@ -513,11 +505,11 @@ class WaveGeom(object):
     #   of 1-3 numbers separated by slashes ('/'). The slashes may
     #   be omitted if there's only one number.
 
-    line = st.getToNextLine()
+    line = st.rval
 
     tabVert = line.split(" ")
 
-    if ((tabVert != None) and (len(tabVert) > 0)):
+    if tabVert:
       for i in range( 0, len(tabVert)):
         sana = tabVert[i]
         posS1 = sana.find('/')
@@ -557,9 +549,8 @@ class WaveGeom(object):
 
             texIdxList.append(texIndex)
 
-    # Add face to current groups
-    self.curGroup.addFace(coordIdxList, texIdxList, normIdxList)
-    # In case we exited early st.skipToNextLine()
+      # Add face to current groups
+      self.curGroup.addFace(coordIdxList, texIdxList, normIdxList)
     # // End of readFace
 
   def readLine(self, st):
@@ -567,7 +558,7 @@ class WaveGeom(object):
     ViewPoint files can have up to two arrays: Vertex Positions, Texture
     Coordinates. Each vertex can contain indices into all two arrays.
     '''
-    count = 0
+  
     vertIndex = 0
     texIndex = 0
 
@@ -576,46 +567,35 @@ class WaveGeom(object):
     # There are n vertices on each line. Each vertex is comprised
     # of 1-2 numbers separated by slashes ('/'). The slashes may
     # be omitted if there's only one number.
-    st.getToken()
+    # st.getToken()
+    line = st.rval
+    tabVert = line.split(" ")
 
-    while (st.ttype != PoserFileParser.TT_EOL):
-      # First token is always a number (or EOL)
-      st.pushBack()
-      st.getNumber()
-      vertIndex = int(st.nval) - 1
-      if (vertIndex < 0):
-        # FIX 20081231 : when reading a file, the real coord list is unknown 
-        # ==> Use the temporary list
-        vertIndex += len(self.coordList) + 1
+    if tabVert:
+      for i in range( 0, len(tabVert)):
+        sana = tabVert[i]
+        posS1 = sana.find('/')
+        vertIndex = int(sana[0:posS1] if (posS1 > 0) else sana) - 1
 
-      coordIdxList.append(vertIndex)
+        if vertIndex < 0: # Relative Index
+          vertIndex += len(self.coordList) + 1
 
-      # Next token is a slash, a number, or EOL. Continue on slash
-      st.getToken()
-      if (st.ttype == ord('/')):
-        # If there's a number after the first slash, read it
-        st.getToken()
-        if (st.ttype == PoserFileParser.TT_WORD):
-          # It's a number
-          st.pushBack()
-          st.getNumber()
-          texIndex = int(st.nval) - 1
-          if (texIndex < 0):
-            # FIX 20081231 : when reading a file, the real coord list is unknown 
-            # ==> Use the temporary list
-            texIndex += len(self.texList) + 1
+        coordIdxList.append(vertIndex)
 
-          texIdxList.append(texIndex)
-          st.getToken()
+        if posS1 > 0:
+          posS2 = sana.find('/', posS1 + 1)
 
-      count+=1
-      # End Of While
+          if posS2 > 0:
+            if (posS2 > posS1 + 1):
+              texIndex = int(sana[posS1+1:posS2]) - 1
+              if (texIndex < 0):
+                texIndex += len(self.texList) + 1
 
-    # Add face to current groups
-    self.curGroup.addLine(coordIdxList, texIdxList)
+              texIdxList.append(texIndex)
+      # End Of for
 
-    # In case we exited early
-    st.skipToNextLine()
+      # Add face to current groups
+      self.curGroup.addLine(coordIdxList, texIdxList)
     # End of readLine
 
   # FIXME: 20200713 def getTexListLength(self): return self.numbTVerts
